--------------------------------------------------------------------------------
-- Copyright (c) 2001 by Trenz Electronic.
-- Duenner Kirchweg 77, 32257 Buende, Germany, www.trenz-electronic.de
--------------------------------------------------------------------------------
-- Project:     TE-XC2S Spartan-II Development System.
-- File:        te-bl.vhd
-- Description: Interface to TE-BL Expansion.
--              *** special version, VGA modified ***
-- History:     2001aug15, FB, created.
--------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;

entity teBL is
	port(
		CLK48:  out STD_LOGIC;                     -- 48MHz clock (USB)
		CLK25:  out STD_LOGIC;                     -- 25MHz clock (VGA)
		PB:     out STD_LOGIC_VECTOR(7 downto 0);  -- push buttons
		SW:     out STD_LOGIC_VECTOR(7 downto 0);  -- dip switches
		LED:    in  STD_LOGIC_VECTOR(7 downto 0);  -- LEDs
		DISP:   in  STD_LOGIC_VECTOR(15 downto 0); -- 7-segment displays
		urxd:   out STD_LOGIC;                     -- USB receive data
		urx0:   out STD_LOGIC;                     -- USB receive SE0 
		utxd:   in  STD_LOGIC;                     -- USB transmit data
		utx0:   in  STD_LOGIC;                     -- USB transmit SE0
		utxoe:  in  STD_LOGIC;                     -- USB transmit output enable
		vr:     in  STD_LOGIC;                     -- VGA red
		vg:     in  STD_LOGIC;                     -- VGA green
		vb:     in  STD_LOGIC;                     -- VGA blue
		vi:     in  STD_LOGIC;                     -- VGA intensity
		vcol:   out STD_LOGIC_VECTOR(9 downto 0);  -- VGA column
		vrow:   out STD_LOGIC_VECTOR(9 downto 0);  -- VGA row
		vvblk:  out STD_LOGIC;                     -- VGA vertical blanking
		vhblk:  out STD_LOGIC;                     -- VGA horizontal blanking
		----------------------------               --
		clk_a1: in  STD_LOGIC;                     -- osc
		iop_a1: out STD_LOGIC;                     -- SEG_E
		ion_a1: out STD_LOGIC;                     -- SEG_D
		iop_a2: out STD_LOGIC;                     -- SEG_C
		ion_a2: out STD_LOGIC;                     -- SEG_DP
		iop_a3: out STD_LOGIC;                     -- SEG_B
		ion_a3: out STD_LOGIC;                     -- SEG_A
		iop_a4: out STD_LOGIC;                     -- SEG_F
		ion_a4: out STD_LOGIC;                     -- SEG_G
		iop_a5: in  STD_LOGIC;                     -- C_PB
		ion_a5: in  STD_LOGIC;                     -- C_SW
		iop_a6: out STD_LOGIC;                     -- MUX_A
		ion_a6: out STD_LOGIC;                     -- MUX_B
		iop_b1: out STD_LOGIC;                     -- MUX_C
		ion_b1: out STD_LOGIC;                     -- VPO
		iop_b2: out STD_LOGIC;                     -- FSE0
		ion_b2: in  STD_LOGIC;                     -- VP
		iop_b3: in  STD_LOGIC;                     -- VM
		ion_b3: out STD_LOGIC;                     -- /OE
		iop_b4: out STD_LOGIC;                     -- /vsync
		ion_b4: out STD_LOGIC;                     -- /hsync
		iop_b5: out STD_LOGIC;                     -- blue
		ion_b5: out STD_LOGIC;                     -- green
		iop_b6: out STD_LOGIC;                     -- red
		ion_b6: out STD_LOGIC                      -- bright
		);
end teBL;

--------------------------------------------------------------------------------
architecture bhv of teBL is
	alias osc    :  STD_LOGIC is clk_a1;  
	alias SEG_E  :  STD_LOGIC is iop_a1;   
	alias SEG_D  :  STD_LOGIC is ion_a1;   
	alias SEG_C  :  STD_LOGIC is iop_a2;   
	alias SEG_DP :  STD_LOGIC is ion_a2;   
	alias SEG_B  :  STD_LOGIC is iop_a3;   
	alias SEG_A  :  STD_LOGIC is ion_a3;   
	alias SEG_F  :  STD_LOGIC is iop_a4;   
	alias SEG_G  :  STD_LOGIC is ion_a4;   
	alias C_PB   :  STD_LOGIC is iop_a5;  
	alias C_SW   :  STD_LOGIC is ion_a5;  
	alias MUX_A  :  STD_LOGIC is iop_a6;   
	alias MUX_B  :  STD_LOGIC is ion_a6;   
	alias MUX_C  :  STD_LOGIC is iop_b1;   
	alias VPO    :  STD_LOGIC is ion_b1;   
	alias FSE0   :  STD_LOGIC is iop_b2;   
	alias VP     :  STD_LOGIC is ion_b2;  
	alias VM     :  STD_LOGIC is iop_b3;  
	alias OE     :  STD_LOGIC is ion_b3;   
	alias vsync  :  STD_LOGIC is iop_b4;   
	alias hsync  :  STD_LOGIC is ion_b4;   
	alias blue   :  STD_LOGIC is iop_b5;   
	alias green  :  STD_LOGIC is ion_b5;   
	alias red    :  STD_LOGIC is iop_b6;   
	alias bright :  STD_LOGIC is ion_b6;
	------------------------------------
	signal mux: UNSIGNED(2 downto 0);
	signal seg: STD_LOGIC_VECTOR(7 downto 0);
	signal disp5, disp6, disp7, disp8: STD_LOGIC_VECTOR(7 downto 0);
	signal pbsel: STD_LOGIC_VECTOR(7 downto 0);
	------------------------------------
	signal clken1: STD_LOGIC;
begin
	
	clk48<= osc;
	
	----------------------------------------------------------------------------
	-- clock divider
	clk_blk: block
		signal div:  UNSIGNED(15 downto 0);
		signal div1a: STD_LOGIC;
		signal div1b: STD_LOGIC;
		begin
		div1a<= div(7); -- 7
		process(osc)
		begin
			if rising_edge(osc) then
				div<= div + 1;
				div1b<= div1a;
				-- synopsys translate_off
				if div(0)= 'U' then div<= (others=> '0'); end if;
				-- synopsys translate_on
			end if;
		end process;
		clken1<= div1a and not(div1b);
	end block;
	
	----------------------------------------------------------------------------
	-- global multiplexer
	mux_blk: block
		signal mux2: UNSIGNED(mux'range);
		begin
		process(osc)
		begin
			if rising_edge(osc) then
				if clken1= '1' then
					mux<= mux + 1;
				end if;
				
				-- synopsys translate_off
				if mux(0)= 'U' then mux<= "000"; end if;
				-- synopsys translate_on
			end if;
		end process;
		
		with mux select seg<=
		LED        when "001", -- LED bar
		disp5      when "010", -- 7-segment U5
		disp6      when "011", -- 7-segment U6
		disp7      when "100", -- 7-segment U7
		disp8      when "101", -- 7-segment U8
		pbsel      when "110", -- push buttons
		LED        when others;
		
		with mux select mux2<=
		"001" when "111" | "000",
		mux   when others;
		
		seg_e <= seg(7);
		seg_d <= seg(6);
		seg_c <= seg(5);
		seg_dp<= seg(4);
		seg_b <= seg(3);
		seg_a <= seg(2);
		seg_f <= seg(1);
		seg_g <= seg(0);
		
		mux_a<= mux2(0);
		mux_b<= mux2(1);
		mux_c<= mux2(2);
	end block;
	
	----------------------------------------------------------------------------
	-- push buttons & switches
	pb_blk: block
		signal pbcnt: UNSIGNED(2 downto 0);
		begin
		
		--------------------------------
		-- button multiplexer
		process(osc)
		begin
			if rising_edge(osc) then
				if clken1= '1' then
					if mux= "000" then
						pbcnt<= pbcnt + 1;
					end if;
					pbsel<= (others=> '0');
					pbsel(TO_INTEGER(pbcnt))<= '1';
				end if;
				
				-- synopsys translate_off
				if pbcnt(0)= 'U' then pbcnt<= "000";          end if;
				if pbsel(0)= 'U' then pbsel<= (others=> '0'); end if;
				-- synopsys translate_on
			end if;
		end process;
		
		--------------------------------
		-- push button register
		pb_gen: for i in 0 to 7 generate
			pb_gen_blk: block
				signal db:        STD_LOGIC_VECTOR(3 downto 0);
				signal pb1, pb2:  STD_LOGIC;
				signal ii_sw:     STD_LOGIC;
				begin
				process(osc)
				begin
					if rising_edge(osc) then
						if clken1= '1' then
							if mux= "110" and TO_INTEGER(UNSIGNED(pbcnt))= i then
								if C_PB= '0' then
									db<= (others=> '0');
								else
									db<= db(db'left-1 downto 0) & '1';
								end if;
							end if;
						end if;
						
						-- synopsys translate_off
						if db(0)= 'U' then db<= (others=> '0'); end if;
						-- synopsys translate_on
					end if;
				end process;
				pb1<= db(db'left);
				pb(i)<= pb1;
				
				--------------------------------
				-- switch register
				process(osc)
				begin
					if rising_edge(osc) then
						if mux= "110" then
							pb2<= pb1;
							if pb1= '1' and pb2= '0' then
								ii_sw<= not(ii_sw);
							end if;
						end if;
						
						-- synopsys translate_off
						if ii_sw= 'U' then ii_sw<= '0'; end if;
						-- synopsys translate_on
					end if;
				end process;
				sw(i)<= ii_sw;
			end block;
		end generate;
	end block;
	
	----------------------------------------------------------------------------
	-- 7-segment displays
	disp_blk: block
		--HEX-to-seven-segment decoder
		-- 
		-- segment encoding		   0 => e 
		--      a                  1 => d 
		--     ---                 2 => c 
		--  f |   | b              3 => dp
		--     ---   <- g          4 => b 
		--  e |   | c              5 => a 
		--     ---  * dp           6 => f 
		--      d                  7 => g
		
		type hex2ledT is array(0 to 15) of STD_LOGIC_VECTOR(7 downto 0);
		constant hex2led: hex2ledT:= (
		"11101110", "00101000", "11001101", "01101101",
		"00101011", "01100111", "11100111", "00101100",
		"11101111", "01101111", "10101111", "11100011",
		"11000001", "11101001", "11000111", "10000111");
		begin
		
		disp5<= hex2led(TO_INTEGER(UNSIGNED(disp(15 downto 12))));
		disp6<= hex2led(TO_INTEGER(UNSIGNED(disp(11 downto  8))));
		disp7<= hex2led(TO_INTEGER(UNSIGNED(disp(7  downto  4))));
		disp8<= hex2led(TO_INTEGER(UNSIGNED(disp(3  downto  0))));
		
	end block;
	
	----------------------------------------------------------------------------
	-- USB transceiver
	usb_blk: block
		signal v: STD_LOGIC_VECTOR(1 downto 0);
		signal u: STD_LOGIC_VECTOR(1 downto 0);
		begin
		
		v<= VP & VM;
		with v select u<=
		"10" when "10",    -- differential one
		"00" when "01",    -- differential zero
		"01" when "00",    -- single-ended zero
		"-0" when others;  -- others
		URXD<= u(u'left);
		URX0<= u(u'right);
		
		-- output
		VPO <= UTXD;
		FSE0<= UTX0;
		OE  <= not(UTXOE);
	end block;
	
	----------------------------------------------------------------------------
	-- vga output
	vga_blk: block
		signal   vclken:                 STD_LOGIC;
		
		signal   r_ii, g_ii,    b_ii:    STD_LOGIC; -- pre-blanking rgb
		signal   blk,  hblk_ii, vblk_ii: STD_LOGIC; -- blanking signals
		
		signal   hcnt:   UNSIGNED(vcol'range);
		signal   vcnt:   UNSIGNED(vrow'range);
		
		--        |__________a     d___          |___________e    h___
		-- rgb   _/__________\_____/___   lines _////////////_____////
		--        |                              |
		--       _|____________b c_____         _|____________f g_____
		-- hsync  |            \_/        vsync  |            \_/
		--
		--        d   = 31.77us                  h   = 16.6ms
		--        c-b =  3.77us                  g-f = 64us
		--        d-c =  1.89us                  h-g =  1.02ms
		--        a   = 25.17us                  e   = 15.25ms
		--        b-a =  0.94us                  f-e =  0.35ms
		-- 
		-- Mode       clock       h         v      a    b    c    d      e   f   g   h
		-- 640x480    25.175MHz   31.5kHz   60Hz   640  664  760  800    480 491 493 525
		-- 640x480    24MHz       31.5kHz   60Hz   640  645  730  762    480 491 493 525 (c-b=90, d-c=45)
		-- 800x600    36MHz       35.1kHz   --     800  824  896  1024   600 601 603 625
		-- 1024x768   65MHz       48.4kHz   --     1024 1032 1176 1344   768 771 777 806
		constant hcnt_a: UNSIGNED(hcnt'range):= TO_UNSIGNED(639, hcnt'length); -- blanking start
		constant hcnt_b: UNSIGNED(hcnt'range):= TO_UNSIGNED(644, hcnt'length); -- h-sync start
		constant hcnt_c: UNSIGNED(hcnt'range):= TO_UNSIGNED(729, hcnt'length); -- h-sync end
		constant hcnt_d: UNSIGNED(hcnt'range):= TO_UNSIGNED(761, hcnt'length); -- h length
		
		constant hcnt_v: UNSIGNED(hcnt'range):= TO_UNSIGNED(699, hcnt'length); -- next line
		
		constant vcnt_e: UNSIGNED(vcnt'range):= TO_UNSIGNED(479, vcnt'length); -- v blanking
		constant vcnt_f: UNSIGNED(vcnt'range):= TO_UNSIGNED(490, vcnt'length); -- v-sync start
		constant vcnt_g: UNSIGNED(vcnt'range):= TO_UNSIGNED(492, vcnt'length); -- v-sync end
		constant vcnt_h: UNSIGNED(vcnt'range):= TO_UNSIGNED(524, vcnt'length); -- v length
		
		begin
		
		------------------------------------
		-- i/o signals
		process(osc)
		begin
			if rising_edge(osc) then
				if vclken= '1' then
					red   <= vr and not(blk);
					green <= vg and not(blk);
					blue  <= vb and not(blk);
					bright<= vi and not(blk);
				end if;
			end if;
		end process;
		
		vcol<= STD_LOGIC_VECTOR(hcnt);
		vrow<= STD_LOGIC_VECTOR(vcnt);
		
		blk  <= hblk_ii or vblk_ii;
		vhblk<= hblk_ii;
		vvblk<= vblk_ii;
		
		------------------------------------
		-- VGA pixel clock
		process(osc)
		begin
			if rising_edge(osc) then
				vclken<= not(vclken);
				
				-- synopsys translate_off
				if vclken= 'U' then vclken<= '0'; end if;
				-- synopsys translate_on
			end if;
		end process;
		--		vclken<= '1';
		clk25<= vclken;
		
		------------------------------------
		-- horizontal timing
		process(osc)
		begin
			if rising_edge(osc) then
				if vclken= '1' then
					if hcnt= hcnt_d then 
						hcnt   <= TO_UNSIGNED(0, hcnt'length);
						hblk_ii<= '0';
					else
						hcnt<= hcnt + 1; 
					end if;
					
					if hcnt= hcnt_a  then hblk_ii<= '1'; end if; -- blanking on
					if hcnt= hcnt_b  then hsync  <= '0'; end if; -- h-sync start
					if hcnt= hcnt_c  then hsync  <= '1'; end if; -- h-sync end
				end if;
				
				-- synopsys translate_off
				if hcnt(0)= 'U' then hcnt   <= TO_UNSIGNED(0, hcnt'length); end if;
				if hblk_ii= 'U' then hblk_ii<= '0';                         end if;
				if hsync= 'U'   then hsync  <= '1';                         end if;
				-- synopsys translate_on
			end if;
		end process;
		
		------------------------------------
		-- vertical timing
		process(osc)
		begin
			if rising_edge(osc) then
				if vclken= '1' then
					if hcnt= hcnt_v then
						if vcnt= vcnt_h then
							vcnt<= TO_UNSIGNED(0, vcnt'length);
							vblk_ii<= '0';
						else
							vcnt<= vcnt + 1;
						end if;
					end if;
					
					if vcnt= vcnt_e then vblk_ii<= '1'; end if; -- blanking on
					if vcnt= vcnt_f then vsync  <= '0'; end if; -- v-sync start
					if vcnt= vcnt_g then vsync  <= '1'; end if; -- v-sync end
				end if;
				
				-- synopsys translate_off
				if vcnt(0)= 'U' then vcnt   <= TO_UNSIGNED(0, vcnt'length); end if;
				if vblk_ii= 'U' then vblk_ii<= '0';                         end if;
				if vsync= 'U'   then vsync  <= '1';                         end if;
				-- synopsys translate_on
			end if;
		end process;
	end block;
	
end architecture;

--------------------------------------------------------------------------------
-- end of file
